/*
 * bgrep -- binary grep: block-oriented search in arbitrary data
 * Copyright (C) 2005  Fabian Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 * bgrep.c -- high-level "core" making use of the various low-level parts
 *
 */

/* include bgrep.h first, so that control macros altering the system headers'
   functionality are defined */
#include "bgrep.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <error.h>

#include "crc32.h"
#include "objects.h"


/* output formats / modes of operation */
int  mode_out_block_offset = 0;
int  mode_out_file_name    = 0;
int  mode_out_matches      = 0;
int  mode_copy_blocks      = 0;
int  mode_hash_gen         = 0;

/* global variables that might be removed / reimplemented in another way later */
int  hash_out_fd;


void out_matches(BGrep_Search *search, char *name)
{
	size_t         match_max_len, matchi;
	BGrep_Matches *matches = search->block_matches;
	BGrep_Match   *match;

	if (matches->valid < 1)
	{
		printf("\nno matching blocks in `%s'.\n\n", name);
		return;
	}

	printf("\nmatching blocks in `%s':\n", name);

	match_max_len = snprintf(NULL, 0, "%lld", bgrep_matches_max(matches));

	for (matchi = 0; matchi < matches->valid; matchi++)
	{
		match = &matches->content[matchi];

		printf("%0*lld*%zd\n",
		       match_max_len, match->value, match->count);
	}

	puts("");
}

void out_match(BGrep_Search *search, BGrep_Pattern *pattern, off_t offset, char *name)
{
	size_t  inner_offset_len;

	if (mode_out_file_name > 0)
		printf("%s:", (name ? name : "(standard input)"));

	/* FIXME: also add offset_len using, e.g., fstat() */
	inner_offset_len = snprintf(NULL, 0, "%zd", search->block_size);

	if (mode_out_block_offset > 0)
		printf("%lld+%0*lld",
		       offset / search->block_size,
		       inner_offset_len,
		       offset % search->block_size);
	else
		printf("%lld", offset);

	/* report what pattern was found */
	/* FIXME: needs escaping */
	printf(":%*s\n", pattern->length, pattern->value);

	/* flush output immediately,
	   as it may take some time until next output & implicit flushing */
	if (fflush(stdout))
		/* FIXME: use another approach for error handling?
		   (like returning and checking the returned value) */
		error(2, errno, "can't write to stdandard output");

	/* remember match for later use */
	/* FIXME: check for errors */
	bgrep_search_add_block_matches(search, pattern, offset);
}

void out_match_hash(BGrep_Search *search, size_t hashi, off_t offset, char *name)
{
	size_t  inner_offset_len, hashes_len;

	/* FIXME: re-unify with out_match() */
	if (mode_out_file_name > 0)
		printf("%s:", (name ? name : "(standard input)"));

	/* FIXME: also add offset_len using, e.g., fstat() */
	inner_offset_len = snprintf(NULL, 0, "%zd", search->block_size);
	hashes_len       = snprintf(NULL, 0, "%zd", search->hashes->size - 1);

	if (mode_out_block_offset > 0)
		printf("%lld+%0*lld",
		       offset / search->block_size,
		       inner_offset_len,
		       offset % search->block_size);
	else
		printf("%lld", offset);

	/* report what hash was found */
	printf("#%0*zd=%08X\n", hashes_len, hashi, search->hashes->content[hashi]);

	/* flush output immediately,
	   as it may take some time until next output & implicit flushing */
	if (fflush(stdout))
		/* FIXME: use another approach for error handling?
		   (like returning and checking the returned value) */
		error(2, errno, "can't write to standard output");

	/* remember match for later use */
	/* FIXME: check for errors */
	bgrep_search_add_block_matches(search, NULL, offset);
}

int copy_blocks(BGrep_Search *search, int fd_in, char *path_base)
{
	int            fd_out, ret = 0;
	ssize_t        valid, written;
	size_t         path_base_len, index_len, match_len, matchi;
	off_t          offset;
	char          *path, *path_index, *write_buffer;
	struct stat    stat_buffer;
	BGrep_Matches *matches;
	BGrep_Match   *match;

	if (search->block_matches->valid < 1)
	{
		printf("\nno matching blocks to copy to output files with base path `%s'.\n\n",
		       path_base);
		return 0;
	}

	printf("\ncopying %zd matching block sequences to output files with base path `%s':\n",
	       search->block_matches->valid, path_base);
	
	/* duplicate block_matches to be able to modify it */
	if ((matches = bgrep_matches_dup(search->block_matches)) == NULL)
	{
		ret = -1;
		goto out;
	}

	path_base_len = strlen(path_base);
	index_len     = snprintf(NULL, 0, "%zd",  matches->valid - 1);
	match_len     = snprintf(NULL, 0, "%lld", bgrep_matches_max(matches));

	/* no "+ 1" needed, as sizeof(...) (hopefully) includes trailing NUL */
	if ((path = malloc((path_base_len + index_len +
	                    sizeof(BGREP_COPY_BLOCKS_SUFFIX))
	                   * sizeof(char))) == NULL)
	{
		error(0, errno, "can't allocate path buffer");
		ret = -1;
		goto out_matches;
	}

	path_index = mempcpy(mempcpy(path, path_base, path_base_len),
	                     BGREP_COPY_BLOCKS_SUFFIX,
	                     sizeof(BGREP_COPY_BLOCKS_SUFFIX) - 1);

	for (matchi = 0; matchi < matches->valid; matchi++)
	{
		match = matches->content + matchi;

		sprintf(path_index, "%0*zd", index_len, matchi + 1);

		printf("%s: %*zd/%zd: %0*lld*%zd\n", path,
		       index_len, matchi + 1, matches->valid,
		       match_len, match->value, match->count);

		/* flush output immediately,
		   as it may take some time until next output & implicit flushing */
		if (fflush(stdout))
		{
			error(0, errno, "can't write to standard output");
			ret = -1;
			goto out_path;
		}

		/* ensure file is not a link */
		if ((lstat(path, &stat_buffer) != -1) || (errno != ENOENT))
		{
			error(0, errno,
			      "can't use existent or invalid output file for match %zd of %zd: `%s'",
			      matchi + 1, matches->valid, path);
			ret = -1;
			goto out_path;
		}

		if ((fd_out = open(path,
		                   O_WRONLY | O_CREAT | O_EXCL,
		                   0644)) == -1)
		{
			error(0, errno,
			      "can't open output file for match %zd of %zd: `%s'",
			      matchi + 1, matches->valid, path);
			ret = -1;
			goto out_path;
		}

		/* position source file pointer */
		offset = match->value * search->block_size;
		if (lseek(fd_in, offset, SEEK_SET) == -1)
		{
			error(0, errno,
			      "can't seek to input file offset %lld for match %zd of %zd",
			      offset, matchi + 1, matches->valid);
			ret = -1;
			goto close;
		}

		/* modifying the matches data is valid
		   as it's only our working copy that is changed */

		/* read & write out again all blocks of this match sequence */
		while (match->count > 0)
		{
			/* read current block */
			/* FIXME: check for / warn about short reads */
			errno = 0;
			if ((valid = read(fd_in,
					  search->block_buffer,
					  search->block_size)) < 1)
			{
				error(0, errno,
				      "can't read from input file at offset %lld for match %zd of %zd",
				      offset, matchi + 1, matches->valid);
				ret = -1;
				goto close;
			}

			/* write block to output file;
			   support partial blocks at end of file */
			write_buffer = search->block_buffer;
			while ((written = write(fd_out, write_buffer, valid)) < valid)
			{
				if (written == -1)
				{
					error(0, errno,
					      "can't write input file block at offset %lld to output file for match %zd of %zd: `%s'",
					      offset, matchi + 1, matches->valid, path);
					ret = -1;
					goto close;
				}

				write_buffer += written;
				valid        -= written;
			}

			/* adjust state data */
			match->value += 1;
			match->count -= 1;
			offset       += search->block_size;
		}

close:
		/* close output file */
		if (TEMP_FAILURE_RETRY(close(fd_out)) == -1)
		{
			error(0, errno,
			      "can't close output file for match %zd of %zd: `%s'",
			      matchi + 1, matches->valid, path);
			ret = -1;
			goto out_path;
		}

		if (ret < 0)
			goto out_path;

		/* indicate count of output files successfully written */
		ret++;
	}

	puts("");

out_path:
	free(path);
out_matches:
	bgrep_matches_free(matches);
out:
	if (ret < 0)
		error(0, 0, "can't copy matching blocks to output files");
	return ret;
}

int hashes_read_fd(BGrep_Search *search, int fd)
{
	struct stat  stat_buffer;
	uint32_t    *dest;
	ssize_t      valid;
	size_t       count;

	if (fstat(fd, &stat_buffer) == -1)
	{
		error(0, errno, "can't get attribute information for hashes file");
		return -1;
	}

	if (!S_ISREG(stat_buffer.st_mode) ||
	    (stat_buffer.st_size % sizeof(uint32_t)) != 0)
	{
		error(0, errno, "invalid hashes file");
		return -1;
	}

	if ((search->hashes = bgrep_hashes_new(stat_buffer.st_size / sizeof(uint32_t))) == NULL)
		return -1;

	dest  = search->hashes->content;
	count = search->hashes->size * sizeof(uint32_t);

	/* read in entire binary hashes file */
	valid = 0;
	while ((count > 0) &&
	       (valid = read(fd, dest, count)) > 0)
	{
		dest  += valid;
		count -= valid;
	}

	/* count still >0 here means either read error or premature end of file */
	if (count > 0)
	{
		bgrep_hashes_free(search->hashes);
		error(0, (valid == -1 ? errno : 0),
		      "can't read in entire hashes file");
		return -1;
	}

	/* FIXME: check for hashes file having been grown during read-in..? */

	return 0;
}

int hashes_read_path(BGrep_Search *search, char *path)
{
	int  fd, ret;

	if ((fd = open(path, O_RDONLY)) == -1)
	{
		error(0, errno, "can't open hashes file `%s'", path);
		return -1;
	}

	if ((ret = hashes_read_fd(search, fd)) < 0)
	{
		error(0, errno, "can't use hashes file `%s'", path);
		ret = -1;
	}

	if (TEMP_FAILURE_RETRY(close(fd)) == -1)
	{
		error(0, errno, "can't close hashes file `%s'", path);
		ret = -1;
	}

	return ret;
}

int hash_write_init(const char *path)
{
	if ((hash_out_fd = open(path, O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1)
	{
		error(0, errno, "can't open hashes output file `%s'", path);
		return -1;
	}

	return 0;
}

int hash_write(uint32_t hash)
{
	ssize_t     written;
	size_t      count = sizeof(uint32_t);
	const char *buf   = (const char *) &hash;

	while ((written = write(hash_out_fd, buf, count)) < count)
	{
		buf   += written;
		count -= written;
	}

	if (written == -1)
	{
		error(0, errno, "can't write to hashes output file");
		return -1;
	}

	return 0;
}

int hash_write_finish()
{
	if (TEMP_FAILURE_RETRY(close(hash_out_fd)) == -1)
	{
		error(0, errno, "can't close hashes output file");
		return -1;
	}

	return 0;
}

int bgrep_fd(BGrep_Search *search, int fd, char *name)
{
	int      ret = 1;
	uint32_t hash;
	size_t   bufi, patsi;
	ssize_t  hashi, valid;
	off_t    offset = 0;

	struct stat    stat_buffer;
	BGrep_Pattern *pat;

	for (patsi = 0; patsi < search->patterns->valid; patsi++)
		search->patterns->content[patsi].index = 0;

	if (fstat(fd, &stat_buffer) == -1)
		error(0, errno, "%s: warning: can't get attribute information -- this may affect warnings about short reads",
		      name);

	if (search->offset_start)
	{
		offset = (search->offset_start / search->block_size) *
		         search->block_size;
		if (lseek(fd, offset, SEEK_SET) == -1)
		{
			error(0, errno, "%s: can't seek to start offset %lld",
			      name, offset);
			return 2;
		}
	}


	while ((valid = read(fd, search->block_buffer, search->block_size)) > 0)
	{
		if (search->offset_end && (search->offset_end <= offset))
			break;

		/* first, compute and check hash */
		if ((mode_hash_gen > 0) || search->hashes)
		{
			/* warn about partially read blocks only in case of
			   hash matching, as this will alter the hash */
			if ((valid < search->block_size) &&
			    (offset + valid < stat_buffer.st_size))
				error(0, 0, "%s: warning: bytes read < block size at offset %lld: %zd < %zd",
				      name, offset, valid, search->block_size);

			hash = crc32(search->block_buffer, valid, 0);

			/* write out computed hash to hashes file */
			if (mode_hash_gen > 0)
				hash_write(hash);

			/* match hash against given hash patterns */
			if (search->hashes)
				if ((hashi = bgrep_hashes_find(search->hashes, hash)) >= 0)
					out_match_hash(search, hashi, offset, name);
		}

		for (bufi = 0; bufi < valid; bufi++)
		{
			for (patsi = 0; patsi < search->patterns->valid; patsi++)
			{
				pat = search->patterns->content + patsi;

				if (search->block_buffer[bufi] == pat->value[pat->index])
				{
					if ((pat->length - pat->index) == 1)
					{
						out_match(search, pat,
						          offset + bufi - pat->length + 1,
						          name);
						ret   = 0;
						pat->index = 0;
					}
					else
						pat->index++;
				}
				else
					pat->index = 0;
			}
		}

		offset += valid;
	}

	if (valid < 0)
	{
		error(0, errno,
		      (name ? "can't read from `%s' after %lld+%lld successful reads" : 
		              "can't read from %s after %lld+%lld successful reads"),
		      (name ? name : "(standard input)"),
		      offset / search->block_size,
		      offset % search->block_size);
		return 2;
	}

	/* compact saved matches */
	/* FIXME: make conditional on whether block_matches still needed? */
	if (bgrep_matches_compact(search->block_matches) < 0)
		return 2;

	if (mode_out_matches > 0)
		out_matches(search, name);

	return ret;
}

int bgrep_path(BGrep_Search *search, char *path)
{
	int  fd, ret;

	if ((fd = open(path, O_RDONLY)) == -1)
	{
		error(0, errno, "can't open `%s'", path);
		return 2;
	}

	ret = bgrep_fd(search, fd, path);

	if (mode_copy_blocks)
		if (copy_blocks(search, fd, path) < 0)
			ret = 2;

	if (TEMP_FAILURE_RETRY(close(fd)) == -1)
	{
		error(0, errno, "can't close `%s'", path);
		ret = 2;
	}

	return ret;
}

int main(int argc, char **argv)
{
	int     opt, argi, ret = 1;
	size_t  block_size = BGREP_BS_DEF;
	char   *tail;

	BGrep_Search *search = NULL;

	while ((opt = getopt(argc, argv, "b:e:k:pBHK:M:OS:V")) != -1)
	{
		switch (opt)
		{
			case 'b':
				if (search)
					error(2, 0,
					      "can't change block size after first pattern");
					/* NOTREACHED */

				errno = 0;
				block_size = strtoul(optarg, &tail, 0);

				if (errno || (*tail != '\0'))
					error(2, errno,
					      "can't convert option argument `%s' to block size",
					      optarg);
					/* NOTREACHED */

				break;

			case 'e':
				if (!search &&
				    ((search = bgrep_search_new(block_size)) == NULL))
					exit(2);
					/* NOTREACHED */

				if (bgrep_patterns_add(search->patterns, strlen(optarg), optarg) < 0)
					exit(2);
					/* NOTREACHED */

				break;

			case 'k':
				if (!search &&
				    ((search = bgrep_search_new(block_size)) == NULL))
					exit(2);
					/* NOTREACHED */

				if (hashes_read_path(search, optarg) < 0)
					exit(2);
					/* NOTREACHED */

				break;

			case 'p':
				mode_out_matches++;
				break;

			case 'B':
				mode_out_block_offset++;
				break;

			case 'H':
				mode_out_file_name++;
				break;

			case 'K':
				mode_hash_gen++;
				if (hash_write_init(optarg) < 0)
					exit(2);
					/* NOTREACHED */
				break;

			case 'M':
				/* FIXME: be verbose about interpretation */
				if (!search)
					if ((search = bgrep_search_new(block_size)) == NULL)
						exit(2);
						/* NOTREACHED */
				errno = 0;
				search->offset_end = strtoul(optarg, &tail, 0);

				if (errno)
					goto err_opt_M;

				if (*tail == '\0')
				{
					search->offset_end += search->offset_start;
					break;
				}

				if (*tail != '+')
					goto err_opt_M;

				search->offset_end *= block_size;
				search->offset_end += strtoul(tail + 1, &tail, 0);

				if (!errno && (*tail == '\0'))
				{
					search->offset_end += search->offset_start;
					break;
				}

err_opt_M:
				error(2, errno, "can't convert option argument `%s' to byte count",
				      optarg);
				/* NOTREACHED */

			case 'O':
				mode_copy_blocks++;
				break;

			case 'S':  /* FIXME: perhaps encapsulate as string -> offset conversion function? */
				/* FIXME: be verbose about interpretation */
				if (!search)
					if ((search = bgrep_search_new(block_size)) == NULL)
						exit(2);
						/* NOTREACHED */
				if (search->offset_end)
					error(2, 0, "can't modify start offset after setting end offset");
					/* NOTREACHED */

				errno = 0;
				search->offset_start = strtoul(optarg, &tail, 0);

				if (errno)
					goto err_opt_S;

				if (*tail == '\0')
					break;

				if (*tail != '+')
					goto err_opt_S;

				search->offset_start *= block_size;
				search->offset_start += strtoul(tail + 1, &tail, 0);

				if (!errno && (*tail == '\0'))
					break;

err_opt_S:
				error(2, errno, "can't convert option argument `%s' to start offset",
				      optarg);
				/* NOTREACHED */

			case 'V':
				printf("\
bgrep v%s -- block-oriented search in arbitrary data\n\
Copyright (C) 2005  Fabian Pietsch\n\n\
This is free software; see the source for copying conditions. There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n",
				       BGREP_VERSION);
				exit(0);

			case '?':
				exit(2);
				/* NOTREACHED */
		}
	}

	if ((argc - optind) == 0)
		error(2, 0, "syntax: PATTERN [FILE [...]]");
		/* NOTREACHED */

	/* no search object prematurely created by pattern option processing? */
	if (!search)
	{
		/* create search object now */
		if ((search = bgrep_search_new(block_size)) == NULL)
			exit(2);
			/* NOTREACHED */

		/* FIXME: has to be adjusted for -S and -M options */
		if (mode_hash_gen <= 0)
		{
			/* no -e options in arguments,
			 * so first non-option argument is the search pattern
			 */
			bgrep_patterns_add(search->patterns,
					   strlen(argv[optind]), argv[optind]);

			/* skip to list of files to search */
			optind++;
		}
	}

	/* no arguments left? */
	if ((argc - optind) == 0)
	{
		if (mode_copy_blocks > 0)
			error(2, 0, "can't copy matching blocks when reading from standard input");

		/* in lack of files, search standard input */
		ret = bgrep_fd(search, STDIN_FILENO, NULL);
	}
	else
	{
		/* more than one argument left? force -H, then */
		if ((argc - optind) > 1)
		{
			mode_out_file_name++;

			if (mode_hash_gen > 0)
				error(2, 0, "can't generate hash file from multiple files");
				/* NOTREACHED */
		}

		/* interpret rest of arguments as files to search */
		for (argi = optind; argi < argc; argi++)
		{
			switch (bgrep_path(search, argv[argi]))
			{
				case 0:
					ret = 0;
					break;

				case 1:
					break;

				default:
					exit(2);
					/* NOTREACHED */
			}

			/* clear matches from this run -- irrelevant/wrong for next file */
			bgrep_matches_resize(search->block_matches, 0);
		}
	}

	if (mode_hash_gen > 0)
		if (hash_write_finish() < 0)
			ret = 2;

	exit(ret);
	/* NOTREACHED */
}
