/*
 * bgrep -- binary grep: block-oriented search in arbitrary data
 * Copyright (C) 2005  Fabian Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 * crc32.c -- CRC32 checksum implementation used to match blocks by their CRC32
 *
 */

/* include bgrep.h first, so that control macros altering the system headers'
   functionality are defined */
#include "bgrep.h"

#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <error.h>

#include "crc32.h"


/* polynomial used for CRC32 lookup table generation */
static const uint32_t  crc32_polynomial = 0xEDB88320UL;

/* CRC32 lookup table; dynamically generated when needed */
static       uint32_t *crc32_table      = NULL;


static int crc32_init()
{
	int  byte, count;

	if (crc32_table)
		return 0;

	if ((crc32_table = malloc(256 * sizeof(uint32_t))) == NULL)
	{
#if 0
		error(0, errno, "can't allocate CRC32 table");
		return -1;
#endif
		/* FIXME: returning -1 might be a valid hash in crc32()...
		   Is there a better way than directly exiting here? */
		error(2, errno, "can't allocate CRC32 table");
		/* NOTREACHED */
	}

	/* compute CRC32 lookup table */
	for (byte = 0; byte < 256; byte++)
	{
		register uint32_t  reg = byte;

		count = 8;
		while (count--)
			if (reg & 0x1UL)
			{
				reg >>= 1;
				reg  ^= crc32_polynomial;
			}
			else
				reg >>= 1;

		crc32_table[byte] = reg;
	}

	return 0;
}

void crc32_deinit()
{
	if (crc32_table)
	{
		free(crc32_table);
		crc32_table = NULL;
	}
}

uint32_t crc32(const void *buf, size_t len, uint32_t reg)
{
	const uint8_t *buf_byte = (const uint8_t *) buf;

	if (!crc32_table)
		crc32_init();

	reg ^= 0xFFFFFFFFUL;
	while (len--)
		reg = (reg >> 8) ^ crc32_table[0xFF & (reg ^ *buf_byte++)];

	return (reg ^ 0xFFFFFFFFUL);
}
