/*
 * bgrep -- binary grep: block-oriented search in arbitrary data
 * Copyright (C) 2005  Fabian Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 * objects.c -- abstraction interface to low level "object" handling
 *
 */

/* include bgrep.h first, so that control macros altering the system headers'
   functionality are defined */
#include "bgrep.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <error.h>
#include <assert.h>

#include "objects.h"


/*
 * BGrep_Pattern
 */

BGrep_Pattern *bgrep_pattern_init(BGrep_Pattern *pattern, size_t length, const char *value)
{
	if ((pattern->value = malloc(length * sizeof(char))) == NULL)
	{
		pattern->value  = NULL;
		pattern->length = 0;
		pattern->index  = 0;
		error(0, errno, "can't initialize a BGrep_Pattern");
		return NULL;
	}

	memcpy(pattern->value, value, length);
	pattern->length = length;
	pattern->index  = 0;
	return pattern;
}

void bgrep_pattern_deinit(BGrep_Pattern *pattern)
{
	free(pattern->value);
}


/*
 * BGrep_Patterns
 */

BGrep_Patterns *bgrep_patterns_new()
{
	BGrep_Patterns *ret;

	if ((ret = calloc(1, sizeof(BGrep_Patterns))) == NULL)
	{
		error(0, errno, "can't allocate a BGrep_Patterns");
		return NULL;
	}

	ret->content = NULL;
	return ret;
}

int bgrep_patterns_resize(BGrep_Patterns *patterns, size_t valid_new)
{
	size_t  patsi = patterns->valid;
	size_t  size_new;

	BGrep_Pattern *content_new;

	while (patsi > valid_new)
		bgrep_pattern_deinit(&patterns->content[--patsi]);

	if (!valid_new)
	{
		free(patterns->content);

		patterns->content = NULL;
		patterns->size    = 0;
		patterns->valid   = 0;
		return 0;
	}

	if (valid_new <= patterns->size)
	{
		patterns->valid = valid_new;
		return 0;
	}

	size_new = (valid_new / BGREP_PATTERNS_BS + 1) * BGREP_PATTERNS_BS;

	if ((content_new = realloc(patterns->content,
	                           size_new * sizeof(BGrep_Pattern))) == NULL)
	{
		error(0, errno, "can't resize a BGrep_Patterns");
		return -1;
	}

	patterns->content = content_new;
	patterns->size    = size_new;
	patterns->valid   = valid_new;
	return 0;
}

int bgrep_patterns_add(BGrep_Patterns *patterns, size_t length, const char *value)
{
	if (bgrep_patterns_resize(patterns, patterns->valid + 1))
		goto out_err;

	if (bgrep_pattern_init(&patterns->content[patterns->valid - 1],
	                       length, value) == NULL)
		goto out_err_resize;

	return 0;

out_err_resize:
	bgrep_patterns_resize(patterns, patterns->valid - 1);
out_err:
	error(0, 0, "can't add pattern to a BGrep_Patterns");
	return -1;
}

void bgrep_patterns_free(BGrep_Patterns *patterns)
{
	int  patsi = patterns->size;

	while (patsi > 0)
		bgrep_pattern_deinit(&patterns->content[--patsi]);
	free(patterns->content);
	free(patterns);
}


/*
 * BGrep_Hashes
 */

BGrep_Hashes *bgrep_hashes_new(size_t size)
{
	BGrep_Hashes *hashes;

	if ((hashes = malloc(sizeof(BGrep_Hashes))) == NULL)
	{
		error(0, errno, "can't allocate a BGrep_Hashes");
		return NULL;
	}

	if ((hashes->content = malloc(size * sizeof(uint32_t))) == NULL)
	{
		free(hashes);
		error(0, errno, "can't allocate a BGrep_Hashes->content");
		return NULL;
	}

	hashes->find_index = 0;
	hashes->size       = size;
	return hashes;
}

size_t bgrep_hashes_find(BGrep_Hashes *hashes, uint32_t hash)
{
	size_t  find_index = hashes->find_index,
	        ret = -1;

	if (!hashes->size)
		return -1;

	do
	{
		if (hashes->content[find_index] == hash)
			ret = find_index;

		find_index++;
		if (find_index >= hashes->size)
			find_index = 0;
	}
	while ((ret == -1) && (find_index != hashes->find_index));

	hashes->find_index = find_index;
	return ret;
}

void bgrep_hashes_free(BGrep_Hashes *hashes)
{
	free(hashes->content);
	free(hashes);
}


/*
 * BGrep_Match
 */

BGrep_Match *bgrep_match_init(BGrep_Match *match, off_t value, size_t count)
{
	match->value = value;
	match->count = count;
	return match;
}

/* for bgrep_match_cmp, see objects.h -- inlined for performance reasons */


/*
 * BGrep_Matches
 */

BGrep_Matches *bgrep_matches_new()
{
	BGrep_Matches *ret;

	if ((ret = calloc(1, sizeof(BGrep_Matches))) == NULL)
	{
		error(0, errno, "can't allocate a BGrep_Matches");
		return NULL;
	}

	ret->content = NULL;
	return ret;
}

int bgrep_matches_resize(BGrep_Matches *matches, size_t valid_new)
{
	size_t       size_new;
	BGrep_Match *content_new;

	if (!valid_new)
	{
		free(matches->content);

		matches->content = NULL;
		matches->size    = 0;
		matches->valid   = 0;
		return 0;
	}

	if (valid_new <= matches->size)
	{
		matches->valid = valid_new;
		return 0;
	}

	size_new = (valid_new / BGREP_MATCHES_BS + 1) * BGREP_MATCHES_BS;

	if ((content_new = realloc(matches->content,
	                           size_new * sizeof(BGrep_Match))) == NULL)
	{
		error(0, errno, "can't resize a BGrep_Matches");
		return -1;
	}

	matches->content = content_new;
	matches->size    = size_new;
	matches->valid   = valid_new;
	return 0;
}

int bgrep_matches_add(BGrep_Matches *matches, off_t value, size_t count)
{
	if (bgrep_matches_resize(matches, matches->valid + 1))
		goto out_err;

	if (bgrep_match_init(&matches->content[matches->valid - 1],
	                     value, count) == NULL)
		goto out_err_resize;

	return 0;

out_err_resize:
	bgrep_matches_resize(matches, matches->valid - 1);
out_err:
	error(0, 0, "can't add match to a BGrep_Matches: `%lld*%zd'", value, count);
	return -1;
}

off_t bgrep_matches_max(const BGrep_Matches *matches)
{
	BGrep_Match *match_max;

	if (matches->valid < 1)
		return -1;

	match_max = &matches->content[matches->valid - 1];
	assert(match_max->count > 0);
	return match_max->value + match_max->count - 1;
}

BGrep_Matches *bgrep_matches_dup(const BGrep_Matches *in)
{
	size_t         in_i;
	BGrep_Matches *out;

	if ((out = bgrep_matches_new()) == NULL)
		goto out_err;

	if (bgrep_matches_resize(out, in->valid))
		goto out_err;

	for (in_i = 0; in_i < in->valid; in_i++)
		/* FIXME: use memcpy()..? */
		out->content[in_i] = in->content[in_i];

	return out;

out_err:
	error(0, 0, "can't duplicate a BGrep_Matches");
	return NULL;
}

int bgrep_matches_compact_cmp(const void *a, const void *b)
{
	return bgrep_match_cmp((const BGrep_Match *) a, (const BGrep_Match *) b);
}

int bgrep_matches_compact(BGrep_Matches *matches)
{
	size_t       value_i, count_new;
	BGrep_Match *writep, *readp, *copyp;

	/* sort the copy's content for easier matches merging */
	qsort(matches->content, matches->valid, sizeof(BGrep_Match),
	      bgrep_matches_compact_cmp);

	/* walk through sorted content and merge consecutive matches
	   to bigger single matches */
	writep = matches->content;
	while (writep < matches->content + matches->valid)
	{
		readp = writep + 1;

		/* for each submatch,
		   check for mergability with following match(es) */
		for (value_i = writep->value;
		     value_i <= writep->value + writep->count;
		     value_i++)
			/* merge in every following match
			   that is starting at this submatch */
			while ((readp < matches->content + matches->valid) &&
			       (value_i == readp->value))
			{
				if ((count_new = value_i - writep->value +
				                 readp->count) > writep->count)
					writep->count = count_new;
				readp++;
			}

		/* remove "null matches" not having been merged above */
		if (!writep->count)
			writep--;

		/* compact's core: move still unexamined matches over those
		   that have been merged or removed */
		/* FIXME: use memmove()..? */
		if (readp > writep + 1)
		{
			for (copyp = writep + 1;
			     readp < matches->content + matches->valid;
			     copyp++, readp++)
				*copyp = *readp;
			matches->valid = copyp - matches->content;
		}

		writep++;
	}

	/* finalize the possibly reduced size,
	   possibly also affecting allocated size */
	if (bgrep_matches_resize(matches, writep - matches->content) < 0)
	{
		error(0, 0, "can't compact a BGrep_Matches");
		return -1;
	}

	return 0;
}

void bgrep_matches_free(BGrep_Matches *matches)
{
	free(matches->content);
	free(matches);
}


/*
 * BGrep_Search
 */

BGrep_Search *bgrep_search_new(size_t block_size)
{
	BGrep_Search *ret;

	if ((ret = calloc(1, sizeof(BGrep_Search))) == NULL)
	{
		error(0, errno, "can't allocate a BGrep_Search");
		return NULL;
	}

	if ((ret->block_buffer = malloc(block_size * sizeof(char))) == NULL)
	{
		error(0, errno, "can't allocate a BGrep_Search->block_buffer");
		goto out_err;
	}

	if ((ret->block_matches = bgrep_matches_new()) == NULL)
		goto out_err_buffer;

	if ((ret->patterns = bgrep_patterns_new()) == NULL)
		goto out_err_matches;

	ret->hashes     = NULL;
	ret->block_size = block_size;
	return ret;

out_err_matches:
	bgrep_matches_free(ret->block_matches);
out_err_buffer:
	free(ret->block_buffer);
out_err:
	free(ret);
	error(0, 0, "can't initialize a BGrep_Search");
	return NULL;
}

int bgrep_search_add_block_matches(BGrep_Search *search, const BGrep_Pattern *pattern, off_t offset)
{
	if (bgrep_matches_add(
	      search->block_matches,
	      offset / search->block_size,
	      pattern ?
	        (offset % search->block_size + pattern->length - 1)
	          / search->block_size + 1 :
	        1
	    ))
		return -1;

	return 0;
}

void bgrep_search_free(BGrep_Search *search)
{
	if (search->hashes)
		bgrep_hashes_free(search->hashes);
	bgrep_patterns_free(search->patterns);
	bgrep_matches_free(search->block_matches);
	free(search->block_buffer);
	free(search);
}
