/*
 * bgrep -- binary grep: block-oriented search in arbitrary data
 * Copyright (C) 2005  Fabian Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 * objects.h -- abstraction interface to low level "object" handling
 *
 */

/* needed for various types used here */
#include <stdlib.h>
#include <stdint.h>

/* pre-allocate space for this many additional objects
   when more space is needed */
#define BGREP_MATCHES_BS  256
#define BGREP_PATTERNS_BS   4


/*
 * high-level objects used by bgrep
 */

typedef struct
{
	size_t  index;
	size_t  length;
	char   *value;
}
BGrep_Pattern;

typedef struct
{
	size_t         size;
	size_t         valid;
	BGrep_Pattern *content;
}
BGrep_Patterns;

typedef struct
{
	size_t    size;
	size_t    find_index;
	uint32_t *content;
}
BGrep_Hashes;

typedef struct
{
	off_t  value;
	size_t count;
}
BGrep_Match;

typedef struct
{
	size_t       size;
	size_t       valid;
	BGrep_Match *content;
}
BGrep_Matches;

typedef struct
{
	off_t           offset_start;
	off_t           offset_end;
	size_t          block_size;
	char           *block_buffer;
	BGrep_Matches  *block_matches;
	BGrep_Patterns *patterns;
	BGrep_Hashes   *hashes;
}
BGrep_Search;


/*
 * prototypes for object handling functions
 */

/* BGrep_Pattern */
BGrep_Pattern *bgrep_pattern_init(BGrep_Pattern *pattern, size_t length, const char *value);
void           bgrep_pattern_deinit(BGrep_Pattern *pattern);

/* BGrep_Patterns */
BGrep_Patterns *bgrep_patterns_new(void);
int             bgrep_patterns_resize(BGrep_Patterns *patterns, size_t valid_new);
int             bgrep_patterns_add(BGrep_Patterns *patterns, size_t length, const char *value);
void            bgrep_patterns_free(BGrep_Patterns *patterns);

/* BGrep_Hashes */
BGrep_Hashes *bgrep_hashes_new(size_t size);
size_t        bgrep_hashes_find(BGrep_Hashes *hashes, uint32_t hash);
void          bgrep_hashes_free(BGrep_Hashes *hashes);

/* BGrep_Match */
BGrep_Match       *bgrep_match_init(BGrep_Match *match, off_t value, size_t count);
static inline int  bgrep_match_cmp(const BGrep_Match *a, const BGrep_Match *b)
{
	int ret = (a->value > b->value) - (a->value < b->value);

	if (!ret)
		ret = (a->count > b->count) - (a->count < b->count);

	return ret;
}

/* BGrep_Matches */
BGrep_Matches *bgrep_matches_new(void);
int            bgrep_matches_resize(BGrep_Matches *matches, size_t valid_new);
int            bgrep_matches_add(BGrep_Matches *matches, off_t value, size_t count);
off_t          bgrep_matches_max(const BGrep_Matches *matches);
BGrep_Matches *bgrep_matches_dup(const BGrep_Matches *in);
int            bgrep_matches_compact_cmp(const void *a, const void *b);
int            bgrep_matches_compact(BGrep_Matches *matches);
void           bgrep_matches_free(BGrep_Matches *matches);

/* BGrep_Search */
BGrep_Search *bgrep_search_new(size_t block_size);
int           bgrep_search_add_block_matches(BGrep_Search *search, const BGrep_Pattern *pattern, off_t offset);
void          bgrep_search_free(BGrep_Search *search);
